---------------------------------------------------------------
--                                                           
--  ADA GNAT INTEGRATED DEVELOPMENT ENVIRONMENT (AdaGIDE)    
--                                                           
--  LEXER.ADB 
--  Description : parses a string into Ada tokens 
--                       
--  By: Dr. Martin C. Carlisle
--                                         
-- LEXER is free software; you can redistribute it and/or
-- modify it without restriction.  However, we ask that you
-- please retain the original author information, and clearly
-- indicate if it has been modified.
--
-- LEXER is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
--                                         
--
---------------------------------------------------------------
WITH Ada.Integer_Text_IO;
WITH Ada.Float_Text_IO;
--WITH Ada.Text_IO;
WITH Ada.Characters.Handling;
WITH Ada.Exceptions;
--WITH Ada.Strings.Fixed;
--WITH Ada.Strings.Maps.Constants;
--WITH Ada.Unchecked_Conversion;


PACKAGE BODY lexer IS



NameLength : CONSTANT Integer := 200; -- see LRM 2.2 (15)
SUBTYPE NameType IS String(1..NameLength);

-- Following is an array of the string corresponding to each reserved word
TYPE ImageArray IS ARRAY(ReservedWords) OF String(1..10);
Images : ImageArray :=
  (abort_t => "ABORT     ",abs_t      => "ABS       ",abstract_t => "ABSTRACT  ",
  accept_t => "ACCEPT    ",access_t   => "ACCESS    ",aliased_t  => "ALIASED   ",
  all_t    => "ALL       ",and_t      => "AND       ",array_t    => "ARRAY     ",
  at_t     => "AT        ",begin_t    => "BEGIN     ",body_t     => "BODY      ",
  case_t   => "CASE      ",constant_t => "CONSTANT  ",declare_t  => "DECLARE   ",
  delay_t  => "DELAY     ",delta_t    => "DELTA     ",digits_t   => "DIGITS    ",
  do_t     => "DO        ",else_t     => "ELSE      ",elsif_t    => "ELSIF     ",
  end_t    => "END       ",entry_t    => "ENTRY     ",exception_t=> "EXCEPTION ",
  exit_t   => "EXIT      ",for_t      => "FOR       ",function_t => "FUNCTION  ",
  generic_t=> "GENERIC   ",goto_t     => "GOTO      ",if_t       => "IF        ",
  in_t     => "IN        ",is_t       => "IS        ",limited_t  => "LIMITED   ",
  loop_t   => "LOOP      ",mod_t      => "MOD       ",new_t      => "NEW       ",
  not_t    => "NOT       ",null_t     => "NULL      ",of_t       => "OF        ",
  or_t     => "OR        ",others_t   => "OTHERS    ",out_t      => "OUT       ",
  package_t=> "PACKAGE   ",pragma_t   => "PRAGMA    ",private_t  => "PRIVATE   ",
  procedure_t=> "PROCEDURE ",protected_t=> "PROTECTED ",raise_t    => "RAISE     ",
  range_t  => "RANGE     ",record_t   => "RECORD    ",rem_t      => "REM       ",
  renames_t=> "RENAMES   ",requeue_t  => "REQUEUE   ",return_t   => "RETURN    ",
  reverse_t=> "REVERSE   ",select_t   => "SELECT    ",separate_t => "SEPARATE  ",
  subtype_t=> "SUBTYPE   ",tagged_t   => "TAGGED    ",task_t     => "TASK      ",
  terminate_t=> "TERMINATE ",then_t     => "THEN      ",type_t     => "TYPE      ",
  until_t  => "UNTIL     ",use_t      => "USE       ",when_t     => "WHEN      ",
  while_t  => "WHILE     ",with_t     => "WITH      ",xor_t      => "XOR       ");

-- This is an array of the lengths of each reserved word
ImageLengthArray : ARRAY(ReservedWords) OF Natural :=
  (abort_t => 5,abs_t      => 3,abstract_t => 8,
  accept_t => 6,access_t   => 6,aliased_t  => 7,
  all_t    => 3,and_t      => 3,array_t    => 5,
  at_t     => 2,begin_t    => 5,body_t     => 4,
  case_t   => 4,constant_t => 8,declare_t  => 7,
  delay_t  => 5,delta_t    => 5,digits_t   => 6,
  do_t     => 2,else_t     => 4,elsif_t    => 5,
  end_t    => 3,entry_t    => 5,exception_t=> 9,
  exit_t   => 4,for_t      => 3,function_t => 8,
  generic_t=> 7,goto_t     => 4,if_t       => 2,
  in_t     => 2,is_t       => 2,limited_t  => 7,
  loop_t   => 4,mod_t      => 3,new_t      => 3,
  not_t    => 3,null_t     => 4,of_t       => 2,
  or_t     => 2,others_t   => 6,out_t      => 3,
  package_t=> 7,pragma_t   => 6,private_t  => 7,
  procedure_t=>9,protected_t=>9,raise_t    => 5,
  range_t  => 5,record_t   => 6,rem_t      => 3,
  renames_t=> 7,requeue_t  => 7,return_t   => 6,
  reverse_t=> 7,select_t   => 6,separate_t => 8,
  subtype_t=> 7,tagged_t   => 6,task_t     => 4,
  terminate_t=>9,then_t    => 4,type_t     => 4,
  until_t  => 5,use_t      => 3,when_t     => 4,
  while_t  => 5,with_t     => 4,xor_t      => 3);

-- This is an array of strings corresponding to each compound delimiter                                           
ImageCompoundDelimiter : ARRAY (CompoundDelimiters) OF String(1..2) :=
  (arrow_t=>"=>",double_dot_t=>"..",double_star_t=>"**",assignment_t=>":=",
   noteq_t=>"/=",geq_t=>">=",
   leq_t=>"<=",left_label_bracket_t=>"<<",right_label_bracket_t=>">>",box_t=>"<>");

-- This is an array of characters corresponding to each delimiter
ImageDelimiter : ARRAY (Delimiters) OF String(1..1) :=
  (ampersand_t=>"&",tick_t=>"'",lparen_t=>"(",rparen_t=>")",times_t=>"*",
   plus_t=>"+",comma_t=>",",minus_t=>"-",dot_t=>".",
   divide_t=>"/",colon_t=>":",semicolon_t=>";",lt_t=>"<",eq_t=>"=",gt_t=>">",
   pipe_t=>"|");


TYPE LookaheadArray IS ARRAY(1..3) OF Character;
LookAhead : LookaheadArray; -- lookahead for GetNextChar
HaveLookAhead : Integer := 0;

InputStringLocation : Integer := 0; -- where are we in string?
InputString : StringPointer;

EndOfString : Exception; -- attempted to get token past end of string
SecondUnget : Exception; -- can't do two ungets in a row

-- used in case we need to unget a CR.
EndOfPreviousLine : Natural;

type Position_Flashback is record
   Column : Natural := 1;
   Line   : Natural := 1;
end record;

-- One token of lookahead
Number_Token_Lookaheads : constant := 2;
Old_Positions : array(1..Number_Token_Lookaheads+1) of Position_Flashback;
TokenAhead : array(1..Number_Token_Lookaheads) of Token;
HaveTokenAhead : Natural := 0;

-- keep track of where we are in the file
CurrentCol : Natural := 1;
CurrentLine : Natural := 1;

FUNCTION GetCurrentLine RETURN Natural IS
BEGIN
  if HaveTokenAhead = 0 then
     RETURN CurrentLine;
  else
     RETURN Old_Positions(Number_Token_Lookaheads+1-HaveTokenAhead).Line;
  end if;
END GetCurrentLine;

------------------------------------------
-- PROCEDURE Initialize
--
-- Sets Line/Col to values or 1,1 (default)
------------------------------------------
PROCEDURE Initialize(Line : IN Integer := 1; Col : IN Integer := 1) IS
BEGIN
  CurrentCol := Col;
  CurrentLine := Line;
END Initialize;

---------------------------
-- FUNCTION StringCompare
--
-- Inputs: two strings
-- Outputs: TRUE if two strings are equal (ignore capitalization)
--
---------------------------
FUNCTION StringCompare(Left,Right : IN String) RETURN Boolean IS
BEGIN
  IF Left'Length /= Right'Length THEN
       RETURN False;
  END IF;

  FOR Index IN Left'RANGE LOOP
    IF Ada.Characters.Handling.To_Upper(Left(Index)) /=
       Ada.Characters.Handling.To_Upper(
          Right(Index-Left'First+Right'First)) THEN
      RETURN False;
    END IF;
  END LOOP;

  RETURN True;
END StringCompare;

--------------------------
-- FUNCTION GetNextChar
--
-- returns next character from input stream or
-- Standard.ASCII.LF on new line
--
-- updates position in file
--------------------------
FUNCTION GetNextChar RETURN Character IS
BEGIN
  InputStringLocation := InputStringLocation + 1; -- update string location

  IF HaveLookAhead > 0 THEN
    IF LookAhead(HaveLookAhead) = Standard.ASCII.LF THEN
      CurrentLine := CurrentLine + 1;
      CurrentCol := 1;
    ELSE
      CurrentCol := CurrentCol+1;
    END IF;
    HaveLookAhead := HaveLookAhead-1;
    RETURN LookAhead(HaveLookAhead+1);
  ELSE

      CurrentCol := CurrentCol + 1;
      IF InputStringLocation-1 > InputString'Last + 1 THEN
        RAISE EndOfString;
      ELSIF InputStringLocation-1 = InputString'Last + 1 OR ELSE
           InputString(InputStringLocation-1) = Standard.ASCII.LF THEN
        EndOfPreviousLine := CurrentCol - 1;
        CurrentLine := CurrentLine + 1;
        CurrentCol := 1;
        RETURN Standard.ASCII.LF;
      ELSE
        RETURN InputString(InputStringLocation-1); -- already incremented
      END IF;
  END IF;
END GetNextChar;

-------------------------
-- PROCEDURE Get_Line
--
-- does same thing as Ada.Text_IO.Get_Line, but from correct stream
-------------------------
PROCEDURE Get_Line (Item : OUT String; Last : OUT Natural) IS
  Start,J : Integer;
BEGIN
    Start := InputStringLocation;
    WHILE GetNextChar /= Standard.ASCII.LF LOOP
      NULL;
    END LOOP;
    J := Item'First;
    -- use -2, one for CR, one b/c location is 1 past always
    FOR I IN Start..InputStringLocation - 2 LOOP
      Item(J) := InputString(I);
      J := J + 1;
    END LOOP;
    Last := Item'First + InputStringLocation-Start-1;
END Get_Line;


-------------------------
-- PROCEDURE UngetChar
--
-- undoes last GetNextChar
-- may not be called twice in succession or
-- raises SecondUngetChar exception
-------------------------
PROCEDURE UngetChar(Item : Character) IS
BEGIN
  InputStringLocation := InputStringLocation - 1; -- update string location

  IF HaveLookAhead = LookAhead'Last THEN
    RAISE SecondUnget;
  ELSE
    IF Item = Standard.ASCII.LF THEN
      CurrentLine := CurrentLine - 1;
      CurrentCol := EndOfPreviousLine;
    ELSE
      CurrentCol := CurrentCol - 1;
    END IF;
    HaveLookAhead := HaveLookAhead + 1;
    LookAhead(HaveLookAhead) := Item;
  END IF;
END UngetChar;

TwoUnderscores: exception;

--------------------------
-- FUNCTION GetComment
--
-- gets rest of line and returns comment token
-- updates Current line and column
--
-- Inputs: none
-- Output: token pointer to comment token
--
-- algorithm:
-- 1) read rest of line
-- 2) create CommentToken
-- 3) return
--------------------------
FUNCTION GetComment RETURN Token IS
  comment : NameType;
  length : Integer;
  startcol,start : Integer;
BEGIN
  start := InputStringLocation - 2;
  startcol := CurrentCol-2;
  Get_Line(Item => comment, Last => length);
  CurrentCol := 1;
  RETURN Token'(     token_type => comment_t,
                                         line => CurrentLine-1,
                                         startcol => startcol,
                                         startloc => start,
                                         endloc => start + length );
END GetComment;

--------------------------
-- PROCEDURE GetStringConstant
--                    
-- Inputs:  first character of string constant '"'
-- Outputs: string literal and its length
--
-- According to LRM, a string literal can consist of characters and "" until a
-- single "
--
-- Algorithm:  
--------------------------
PROCEDURE GetStringConstant (FirstChar : IN Character;
                   Name : OUT String;
                   Length : OUT Positive) IS

  NextChar : Character;
  IllegalString : exception;
BEGIN
  Name(1) := FirstChar;
  NextChar := GetNextChar;
  Length := 2;

  LOOP
    WHILE (NextChar /= '"') LOOP
      IF NextChar = Standard.ASCII.LF THEN
        UngetChar(NextChar);
        RAISE IllegalString;
      END IF;
      Name(Length) := NextChar;
      Length := Length+1;
      NextChar := GetNextChar;
    END LOOP;
    Name(Length) := '"';
    Length := Length + 1;
    NextChar:=GetNextChar;
    EXIT WHEN NextChar /= '"';
    NextChar:=GetNextChar;
    Length := Length + 1;
  END LOOP;

  -- we got one more character than we need
  UngetChar(NextChar);
  Length := Length - 1;

END GetStringConstant;

--------------------------
-- PROCEDURE GetIdentifier
--
-- Inputs:  first character of identifier
-- Outputs: identifier and its length
--
-- According to LRM, an identifier may consist of a letter followed by
-- a sequence of letters, digits and underscores.  (However, there may not be
-- two adjacent underscores).
--
-- Algorithm:  this procedure loops until it finds the end of the identifier,
--             and returns the identifier and its length
-- exceptions: TwoUnderscores
--------------------------
PROCEDURE GetIdentifier (FirstChar : IN Character;
                   Name : OUT String;
                   Length : OUT Positive) IS

  NextChar : Character;
  WasUnderscore : Boolean := FALSE;

BEGIN
  NextChar := FirstChar;
  Length := 1;

  WHILE (Ada.Characters.Handling.Is_Letter(NextChar) OR
         NextChar = '_' OR
         Ada.Characters.Handling.Is_Digit(NextChar)) LOOP
    WasUnderscore := (NextChar = '_');
    Name(Length) := NextChar;
    Length := Length+1;
    NextChar := GetNextChar;
--    IF WasUnderscore AND NextChar = '_' THEN
--      RAISE TwoUnderscores;
--    END IF;
  END LOOP;

  -- we got one more character than we need
  UngetChar(NextChar);
  Length := Length -1;
END GetIdentifier;

ErrorInNumber : exception;
------------------------------------------
-- GetNumber 
--
-- reads either an integer constant or floating point
-- constant
--
-- inputs: first digit
-- output: either integer or floating result and 
--         boolean saying whether a decimal point was found 
--
-- algorithm:
-- 1) get integer part
-- 2) check for decimal point
--   a) if not, check for #
--      if #, then read through to first blank, checking for .
--        if . found, use float_text_io.get o/w integer_text_io.get
--        trap exception and report error
--      else
--        compute IntValue
--   b) else, ensure one digit
--     2) get decimal part
--     3) check for exponent
--       a) if present
--         1) get sign if present and ensure one digit
--         2) get rest of exponent
--     4) compute FloatResult
--
-- exceptions: ErrorInNumber
------------------------------------------
PROCEDURE GetNumber (FirstDigit : IN Character;
                     IntResult  : OUT Integer;
                     FloatResult : OUT Float;
                     HasDecimal  : OUT Boolean;
                     endloc : OUT Positive) IS


  NumString : NameType;
  NextChar : Character;
  Counter : Integer;
  Last : Integer;

  PROCEDURE GetExponent IS
  BEGIN
    IF NextChar = 'E' OR NextChar = 'e' THEN
      NumString(Counter) := 'E';
      Counter := Counter + 1;
      NextChar := GetNextChar;

      IF NextChar = '+' OR NextChar = '-' THEN
        NumString(Counter) := NextChar;
        Counter := Counter + 1;
        NextChar := GetNextChar;
      END IF;

      IF NOT Ada.Characters.Handling.Is_Digit(NextChar) THEN
        UngetChar(NextChar);
        Ada.Exceptions.Raise_Exception(ErrorInNumber'Identity,
          "Must have at least one in exponent");
      END IF;

      WHILE (Ada.Characters.Handling.Is_Digit(NextChar)OR
             NextChar = '_') LOOP
        NumString(Counter) := NextChar;
        Counter := Counter + 1;
        NextChar := GetNextChar;
      END LOOP;
    END IF;
  END GetExponent;

BEGIN
  HasDecimal := FALSE;
  NextChar := FirstDigit;
  Counter := 1;
  WHILE (Ada.Characters.Handling.Is_Digit(NextChar) OR
         NextChar = '_') LOOP
    NumString(Counter) := NextChar;
    Counter := Counter + 1;
    NextChar := GetNextChar;
  END LOOP;

  IF NextChar /= '.' THEN
    IF NextChar = '#' THEN
      WHILE Ada.Characters.Handling.Is_Digit(NextChar) OR ELSE
            Ada.Characters.Handling.Is_Letter(NextChar) OR ELSE
            NextChar = '_' OR ELSE NextChar = '.' OR ELSE
            NextChar = '#' LOOP
        IF NextChar = '.' THEN
          HasDecimal := TRUE;
        END IF;
        NumString(Counter) := NextChar;
        Counter := Counter + 1;
        NextChar := GetNextChar;
      END LOOP;
      GetExponent;
      UngetChar(NextChar);
      IF HasDecimal THEN
        Ada.Float_Text_IO.Get(From => NumString(1..Counter-1), Item => FloatResult,
                            Last => Last);
      ELSE
        Ada.Integer_Text_IO.Get(From => NumString(1..Counter-1), Item => IntResult,
                                Last => Last);
      END IF;
    ELSE
      GetExponent;
      UngetChar(NextChar);
      Ada.Integer_Text_IO.Get(From => NumString(1..Counter-1), Item => IntResult,
                              Last => Last);
      HasDecimal := FALSE;
    END IF;
  ELSE
    NextChar := GetNextChar;
    IF NextChar = '.' THEN -- .. token should be added for next time
      HasDecimal := FALSE;
      Ada.Integer_Text_IO.Get(From => NumString(1..Counter-1),
         Item => IntResult, Last => Last);
      endloc := InputStringLocation - 3;
      UngetChar('.');
      UngetChar('.');
      RETURN;
    END IF;
    HasDecimal := TRUE;
    NumString(Counter) := '.';
    Counter := Counter + 1;

    IF NOT Ada.Characters.Handling.Is_Digit(NextChar) THEN
      HasDecimal := FALSE;
      UngetChar(NextChar);
      UngetChar('.');
      Counter := Counter -1;
      Ada.Integer_Text_IO.Get(From => NumString(1..Counter-1),
         Item => IntResult, Last => Last);
      endloc := InputStringLocation - 3;
      RETURN;
      -- 7/2/99 (mcc)
      -- modified for colorization so we can type 10..5 without having
      -- an error after just 10. 
      -- similar to .. above
      --
      -- old error is below
      --Ada.Exceptions.Raise_Exception(ErrorInNumber'Identity,
      --  "Must have at least one digit after decimal point");
    END IF;

    WHILE (Ada.Characters.Handling.Is_Digit(NextChar)OR
           NextChar = '_') LOOP
      NumString(Counter) := NextChar;
      Counter := Counter + 1;
      NextChar := GetNextChar;
    END LOOP;

    GetExponent;
    UngetChar(NextChar);
    Ada.Float_Text_IO.Get(From => NumString(1..Counter-1), Item => FloatResult,
                            Last => Last);

  END IF;

  endloc := InputStringLocation - 1;
EXCEPTION
  -- this should happen if a IO.Get call fails
  WHEN others =>
    Ada.Exceptions.Raise_Exception(ErrorInNumber'Identity,
          NumString(1..Counter-1) & " is not a valid number");
    endloc := InputStringLocation - 1;
END GetNumber;




--------------------------------
-- PROCEDURE UngetToken
--
-- undoes last GetToken
-- cannot be called twice in succession, or 
-- exception CannotUnget is raised
--------------------------------
PROCEDURE UngetToken(tok : Token) IS
BEGIN
  IF HaveTokenAhead=Number_Token_Lookaheads THEN
    RAISE CannotUnget;
  ELSE
    HaveTokenAhead := HaveTokenAhead+1;
    TokenAhead(HaveTokenAhead) := tok;
  END IF;
END UngetToken;

------------------------------
-- FUNCTION CheckReservedWords
--
-- Inputs : a string, its length and first character
-- Output : a token pointer of the appropriate reserved word or
--          identifier
------------------------------
FUNCTION CheckReservedWords (Name : IN String;
                             Length : IN Natural;
                             FirstChar : IN Character) RETURN Token IS
  First,Last : ReservedWords;
BEGIN
  CASE FirstChar IS
    WHEN 'A'|'a' => first := abort_t; last := at_t;
    WHEN 'B'|'b' => first := begin_t; last := body_t;
    WHEN 'C'|'c' => first := case_t; last := constant_t;
    WHEN 'D'|'d' => first := declare_t; last := do_t;
    WHEN 'E'|'e' => first := else_t; last := exit_t;
    WHEN 'F'|'f' => first := for_t; last := function_t;
    WHEN 'G'|'g' => first := generic_t; last := goto_t;
    WHEN 'I'|'i' => first := if_t; last := is_t;
    WHEN 'L'|'l' => first := limited_t; last := loop_t;
    WHEN 'M'|'m' => first := mod_t; last := mod_t;
    WHEN 'N'|'n' => first := new_t; last := null_t;
    WHEN 'O'|'o' => first := of_t; last := out_t;
    WHEN 'P'|'p' => first := package_t; last := protected_t;
    WHEN 'R'|'r' => first := raise_t; last := reverse_t;
    WHEN 'S'|'s' => first := select_t; last := subtype_t;
    WHEN 'T'|'t' => first := tagged_t; last := type_t;
    WHEN 'U'|'u' => first := until_t; last := use_t;
    WHEN 'W'|'w' => first := when_t; last := with_t;
    WHEN 'X'|'x' => first := xor_t; last := xor_t;
    WHEN others => first := xor_t; last := abort_t; -- null range
  END CASE;

  -- check against all reserved words beginning with same letter as Name
  FOR Index IN first..last LOOP
    IF Length = ImageLengthArray(Index) AND THEN
       StringCompare(Images(Index)(1..Length),Name(1..Length)) THEN
      RETURN Token'(token_type => Index,
                    startcol => CurrentCol - Length,
                    startloc => InputStringLocation-Length,
                    endloc => InputStringLocation-1,
                    line => CurrentLine);
    END IF;
  END LOOP;

  -- not a reserved word
  RETURN Token'(token_type => name_t,
                startcol => CurrentCol-Length,
                startloc => InputStringLocation-Length,
                endloc => InputStringLocation-1,
                line => CurrentLine);

END CheckReservedWords;

--------------------------------------------
-- PROCEDURE GetNextToken
--
-- the heart of the lexer, returns a pointer 
-- to the next token object in input stream
--
-- Algorithm:
--    This is a simple LL(1) descent lexer, which 
--    checks the first character, and then 
--    branches to an appropriate handler
--    If the first character is not sufficient to 
--    identify the token type, a nested IF is used 
-- exceptions:
--   BadCharacter
---------------------------------------------
BadCharacter : Exception;
PROCEDURE GetNextToken (Str : IN StringPointer;
  First : IN Integer; Last : OUT Integer;
  Result : OUT Token;
  HaveOne : OUT Boolean; Done : OUT Boolean) IS

  NextChar : Character;
  IntValue : Integer;
  FloatValue : Float;
  IsFloat : Boolean;
  Name : NameType;
  Length : Integer;
  OldCol,OldLoc,endloc : Positive;
BEGIN
  Done := False;
  HaveOne := True;
  HaveLookAhead := 0; -- only since reading from a string
  InputString := Str;


  IF HaveTokenAhead > 0 THEN
    Result :=  TokenAhead(HaveTokenAhead);
    Last := TokenAhead(HaveTokenAhead).endloc;
--    CurrentLine := Old_Positions(
--       Number_Token_Lookaheads+1-HaveTokenAhead).Line;
--    CurrentCol  := Old_Positions(
--       Number_Token_Lookaheads+1-HaveTokenAhead).Column;
    HaveTokenAhead := HaveTokenAhead - 1;
    RETURN;
  END IF;

  InputStringLocation := First;
  Last := InputStringLocation;   -- important in case of exception
  NextChar := GetNextChar;

  -- skip white space
  WHILE Ada.Characters.Handling.Is_Control(NextChar) OR ELSE
        NextChar = ' ' LOOP
    NextChar := GetNextChar;
  END LOOP;

  -- used in case of error
  OldCol := CurrentCol-1;
  OldLoc := InputStringLocation-1;

  CASE NextChar IS
    WHEN '&' => Result := Token'(token_type => ampersand_t,
                                 line => CurrentLine,
                                 startloc => InputStringLocation-1,
                                 startcol => CurrentCol-1,
                                 endloc => InputStringLocation-1);
    WHEN ''' =>
       DECLARE
          AfterTick,TwoAfterTick : Character;
       BEGIN
          AfterTick := GetNextChar;
          TwoAfterTick := GetNextChar;
          IF TwoAfterTick /= ''' THEN
            UngetChar(TwoAfterTick);
            UngetChar(AfterTick);
            Result := Token'(token_type => tick_t,
                                 line => CurrentLine,
                                 startcol => CurrentCol-1,
                                 startloc => InputStringLocation-1,
                                 endloc => InputStringLocation-1);
          ELSE
            Result := Token'(token_type => char_literal_t,
                                 line => CurrentLine,
                                 startcol => CurrentCol-3,
                                 startloc => InputStringLocation-3,
                                 endloc => InputStringLocation-1);
          END IF;
       END;
    WHEN '(' => Result := Token'(token_type => lparen_t,
                                 line => CurrentLine,
                                 startcol => CurrentCol-1,
                                 startloc => InputStringLocation-1,
                                 endloc => InputStringLocation-1);
    WHEN ')' => Result := Token'(token_type => rparen_t,
                                 line => CurrentLine,
                                 startcol => CurrentCol-1,
                                 startloc => InputStringLocation-1,
                                 endloc => InputStringLocation-1);
    WHEN '*' => NextChar := GetNextChar;
                IF NextChar = '*' THEN
                  Result :=  Token'(token_type => double_star_t,
                                    line => CurrentLine,
                                    startcol => CurrentCol-2,
                                    startloc => InputStringLocation-2,
                                    endloc => InputStringLocation-1);
                ELSE
                  UnGetChar(Item => NextChar);
                  Result :=  Token'(token_type => times_t,
                                    line => CurrentLine,
                                    startcol => CurrentCol-1,
                                    startloc => InputStringLocation-1,
                                    endloc => InputStringLocation-1);
                END IF;
    WHEN '+' => Result :=  Token'(token_type => plus_t,
                                  line => CurrentLine,
                                  startcol => CurrentCol-1,
                                  startloc => InputStringLocation-1,
                                  endloc => InputStringLocation-1);
    WHEN ',' => Result :=  Token'(token_type => comma_t,
                                  line => CurrentLine,
                                  startcol => CurrentCol-1,
                                  startloc => InputStringLocation-1,
                                  endloc => InputStringLocation-1);
    WHEN '-' => NextChar := GetNextChar;
                IF NextChar = '-' THEN
                  Result :=  GetComment;
                ELSE
                  UnGetChar(Item => NextChar);
                  Result :=  Token'(token_type => minus_t,
                                    line => CurrentLine,
                                    startcol => CurrentCol-1,
                                    startloc => InputStringLocation-1,
                                    endloc => InputStringLocation-1);
                END IF;
    WHEN '.' => NextChar := GetNextChar;
                IF NextChar = '.' THEN
                  Result :=  Token'(token_type => double_dot_t,
                                    line => CurrentLine,
                                    startcol => CurrentCol-2,
                                    startloc => InputStringLocation-2,
                                    endloc => InputStringLocation-1);
                ELSE
                  UnGetChar(Item => NextChar);
                  Result :=  Token'(token_type => dot_t,
                                    startcol => CurrentCol-1,
                                    line => CurrentLine,
                                    startloc => InputStringLocation-1,
                                    endloc => InputStringLocation-1);
                END IF;
    WHEN '/' => NextChar := GetNextChar;
                IF NextChar = '=' THEN
                  Result :=  Token'(token_type => noteq_t,
                                    line => CurrentLine,
                                    startcol => CurrentCol-2,
                                    startloc => InputStringLocation-2,
                                    endloc => InputStringLocation-1);
                ELSE
                  UnGetChar(Item => NextChar);
                  Result :=  Token'(token_type => divide_t,
                                    line => CurrentLine,
                                    startcol => CurrentCol-1,
                                    startloc => InputStringLocation-1,
                                    endloc => InputStringLocation-1);
                END IF;
    WHEN ':' => NextChar := GetNextChar;
                IF NextChar = '=' THEN
                  Result :=  Token'(token_type => assignment_t,
                                    line => CurrentLine,
                                    startcol => CurrentCol-2,
                                    startloc => InputStringLocation-2,
                                    endloc => InputStringLocation-1);
                ELSE
                  UnGetChar(Item => NextChar);
                  Result :=  Token'(token_type => colon_t,
                                    line => CurrentLine,
                                    startcol => CurrentCol-1,
                                    startloc => InputStringLocation-1,
                                    endloc => InputStringLocation-1);
                END IF;
    WHEN ';' => Result :=  Token'(token_type => semicolon_t,
                                  line => CurrentLine,
                                  startcol => CurrentCol-1,
                                  startloc => InputStringLocation-1,
                                  endloc => InputStringLocation-1);
    WHEN '<' => NextChar := GetNextChar;
                IF NextChar = '=' THEN
                  Result :=  Token'(token_type => leq_t,
                                    line => CurrentLine,
                                    startcol => CurrentCol-2,
                                    startloc => InputStringLocation-2,
                                    endloc => InputStringLocation-1);
                ELSIF NextChar = '<' THEN
                  Result := Token'(token_type => left_label_bracket_t,
                                   line => CurrentLine,
                                   startcol => CurrentCol-2,
                                   startloc => InputStringLocation-2,
                                   endloc => InputStringLocation-1);
                ELSIF NextChar = '>' THEN
                  Result :=  Token'(token_type => box_t,
                                    line => CurrentLine,
                                    startcol => CurrentCol-2,
                                    startloc => InputStringLocation-2,
                                    endloc => InputStringLocation-1);
                ELSE
                  UnGetChar(Item => NextChar);
                  Result :=  Token'(token_type => lt_t,
                                    line => CurrentLine,
                                    startcol => CurrentCol-2,
                                    startloc => InputStringLocation-1,
                                    endloc => InputStringLocation-1);
                END IF;
    WHEN '=' => NextChar := GetNextChar;
                IF NextChar = '>' THEN
                  Result :=  Token'(token_type => arrow_t,
                                    line => CurrentLine,
                                    startcol => CurrentCol-2,
                                    startloc => InputStringLocation-2,
                                    endloc => InputStringLocation-1);
                ELSE
                  UnGetChar(Item => NextChar);
                  Result :=  Token'(token_type => eq_t,
                                    line => CurrentLine,
                                    startcol => CurrentCol-1,
                                    startloc => InputStringLocation-1,
                                    endloc => InputStringLocation-1);
                END IF;
    WHEN '>' => NextChar := GetNextChar;
                IF NextChar = '=' THEN
                  Result :=  Token'(token_type => geq_t,
                                    line => CurrentLine,
                                    startcol => CurrentCol-2,
                                    startloc => InputStringLocation-2,
                                    endloc => InputStringLocation-1);
                ELSIF NextChar = '>' THEN
                  Result :=  Token'(token_type => right_label_bracket_t,
                                    line => CurrentLine,
                                    startcol => CurrentCol-2,
                                    startloc => InputStringLocation-2,
                                    endloc => InputStringLocation-1);
                ELSE
                  UnGetChar(Item => NextChar);
                  Result :=  Token'(token_type => gt_t,
                                    line => CurrentLine,
                                    startcol => CurrentCol-1,
                                    startloc => InputStringLocation-1,
                                    endloc => InputStringLocation-1);
                END IF;
    WHEN '|' => Result :=  Token'(token_type => pipe_t,
                                  line => CurrentLine,
                                  startcol => CurrentCol-1,
                                  startloc => InputStringLocation-1,
                                  endloc => InputStringLocation-1);
    WHEN '0'..'9' =>
            GetNumber(FirstDigit => NextChar,
                      IntResult =>  IntValue,
                      FloatResult => FloatValue,
                      HasDecimal => IsFloat,
                      endloc => endloc);
            IF IsFloat THEN
              Result :=  Token'(token_type=>float_t,
                                endloc => endloc,
                                startcol => OldCol,
                                startloc => OldLoc,
                                line => CurrentLine);
            ELSE
              Result :=  Token'(token_type=>integer_t,
                                endloc => endloc,
                                startcol => OldCol,
                                startloc => OldLoc,
                                line => CurrentLine);
            END IF;
    WHEN '"' =>
      GetStringConstant(FirstChar => NextChar,
                        Name => Name,
                        Length => Length);
      Result := Token'(token_type=>string_t,
                       line=>CurrentLine,
                       startcol => CurrentCol-Length,
                       startloc=>InputStringLocation-Length,
                       endloc=>InputStringLocation-1);
    WHEN 'A'..'Z'|'a'..'z' =>
      GetIdentifier(FirstChar => NextChar,
                    Name => Name,
                    Length => Length);
      Result :=  CheckReservedWords(Name => Name(1..Length), Length => Length,
                                FirstChar => NextChar);
    WHEN OTHERS =>
       IF Ada.Characters.Handling.Is_Letter(NextChar) THEN
          GetIdentifier(FirstChar => NextChar,
             Name   => Name,
             Length => Length);
          Result :=  CheckReservedWords(Name => Name(1..Length),
             Length    => Length,
             FirstChar => NextChar);
       ELSE
          Ada.Exceptions.Raise_Exception(BadCharacter'Identity,
             "Character is not valid in Ada");
       END IF;
  END CASE;
  Last := InputStringLocation-1;
  for i in reverse Number_Token_Lookaheads+1..2 loop
     Old_Positions(i) := Old_Positions(i-1);
  end loop;
  Old_Positions(1).Column := CurrentCol;
  Old_Positions(1).Line   := CurrentLine;
EXCEPTION
  WHEN EndOfString =>
    HaveOne := False;
    Done := True;
  WHEN E: OTHERS =>
    Result := Token'(token_type => error_t,
                       line=>CurrentLine,
                       startcol => OldCol,
                       startloc=> Oldloc,
                       endloc=>InputStringLocation-1);
    HaveOne := True;
    Last := InputStringLocation-1;
END GetNextToken;


FUNCTION GetTokenClass(Item : Token) RETURN TokenClass IS
BEGIN
  IF Item.token_type IN ReservedWords THEN
    RETURN ReservedClass;
  ELSIF Item.token_type IN Constants THEN
    RETURN ConstantClass;
  ELSIF Item.token_type = string_t THEN
    RETURN StringClass;
  ELSIF Item.token_type = comment_t THEN
    RETURN CommentClass;
  ELSIF Item.token_type = name_t THEN
    RETURN IdentifierClass;
  ELSE
    RETURN OtherClass;
  END IF;
EXCEPTION
  WHEN OTHERS =>
    RETURN OtherClass;
END GetTokenClass;

FUNCTION GetTokenType(Item : Token) RETURN Tokens IS
BEGIN
  RETURN Item.token_type;
END GetTokenType;

FUNCTION GetStart(Item : Token) RETURN Integer IS
BEGIN
  RETURN Item.startloc;
END GetStart;

FUNCTION GetEnd(Item : Token) RETURN Integer IS
BEGIN
  RETURN Item.endloc;
END GetEnd;

FUNCTION GetCol(Item : Token) RETURN Integer IS
BEGIN
  RETURN Item.startcol;
END GetCol;

FUNCTION GetLine(Item : Token) RETURN Integer IS
BEGIN
  RETURN Item.line;
END GetLine;


END lexer;